@@ -1,4 +1,4 @@
-# Copyright (C) 2005 by Johan Dahlin
+# Copyright (C) 2005-2006 by Johan Dahlin
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public License
@@ -16,16 +16,15 @@
 
 # properties.py - More black magic than Dumbledore can handle
 
-import gettext
 import sys
 
 import gobject
 import gtk
+from kiwi.environ import environ
 
 from gazpacho.choice import enum_to_string, flags_to_string
-from gazpacho.environ import environ
-
-_ = gettext.gettext
+from gazpacho.command import Command
+from gazpacho.i18n import _
 
 UNICHAR_PROPERTIES = ('invisible-char', )
 
@@ -35,12 +34,19 @@ class PropertyError(Exception):
 class UnsupportedProperty(PropertyError):
     pass
 
+class PropertySetError(PropertyError):
+    """
+    This can only be raised inside a set handler of a PropType subclass.
+    It is raised when an invalid value set, for example when the position
+    of a cell in a table is outside of the table borders.
+    """
+
 class PropertyCustomEditor(object):
     """
     Base class for editors for custom properties
 
     The create method is called the first time a widget with this property is
-    shown in the editor. Then, everytime a instance of this type of widget is
+    shown in the editor. Then, every time an instance of this type of widget is
     selected the 'update' method is called with the proper information to
     manipulate the widget.
 
@@ -66,208 +72,18 @@ class PropertyCustomEditor(object):
                            works
         """
 
-class PropRegistry:
+class PropStorage:
+    """A PropStorage is a bag where we keep PropTypes
+
+    It is used in PropRegistry to store the regular properties and the child
+    properties.
+    """
+    child = False
     def __init__(self):
-        self._prop_types = {}
-        self._bases = {}
-
-        # A list of custom properties
+        self._types = {}
         self._custom = {}
-
-        # Cache used by _list_properties, list of pspecs
         self._pspec_cache = {}
 
-    def add_base_type(self, klass, gtype):
-        if not issubclass(klass, PropType):
-            raise TypeError("klass needs to be a subclass of PropType")
-
-        if gtype in self._bases:
-            raise PropertyError("base type %r is already defined" % gtype)
-        elif klass.base_type != None:
-            raise PropertyError("klass type %r is already registered to %r" %
-                                (klass, klass.base_type))
-
-        self._bases[gtype] = klass
-        klass.base_type = gtype
-
-    def override_property(self, full_name, klass, child=False):
-        """
-        Create a property for full_name, (type::prop_name)
-        If it doesn't exist add a custom property
-        @param klass: property class
-        @type  klass: subclass of PropType
-        @param full_name: complete name of property
-        @type  full_name: string (typename::prop_name)
-        """
-
-        if not issubclass(klass, PropType):
-            raise TypeError("klass needs to be a subclass of PropType")
-
-        type_name, prop_name = full_name.split('::')
-        pspec = self._get_pspec(type_name, prop_name)
-
-        # If we can't find the property in the GType, it's a custom type
-        if not self._has_property(type_name, prop_name):
-            self._add_custom_type(klass, type_name, prop_name, pspec)
-        else:
-            # XXX: Kill this, needs to fix readable/writable first
-            #raise TypeError("%s exists, use override_simple" % full_name)
-            self._set_interval_from_pspec(klass, pspec)
-            self._add_type(full_name, klass, pspec)
-
-        # Note, that at this point we need to setup the overriden class
-        # since we're not going to call the type constructor, so
-        # this needs to be in sync with PropMeta.new
-
-        klass.name = prop_name
-        klass.owner_name = type_name
-        klass.owner_type = gobject.type_from_name(type_name)
-
-        # Do not always override, so a base type can specify it
-        if klass.child is None:
-            klass.child = child
-
-        if not issubclass(klass.editor, PropertyCustomEditor):
-            raise TypeError("editor %r for property class %r needs to be a "
-                            "subclass of PropertyCustomEditor " % (
-                klass.editor, klass))
-
-    def override_simple(self, full_name, base=None, child=False, **kwargs):
-        """
-        Same as override_property, but creates and registeres the class
-        @param full_name: complete name of property
-        @param base:      base class
-        @param child:     child property
-        @param kwargs namespace to use in class
-        """
-
-        type_name, prop_name = full_name.split('::')
-        parent = None
-        if child or (base and base.child):
-            parent = type_name
-
-        pspec = self._get_pspec(type_name, prop_name, parent)
-        if not pspec:
-            raise TypeError("You can only override existing properties, %s "
-                            "is not a property" % full_name)
-
-        if not base:
-            base = CustomProperty
-
-        parent = self._get_base_type(pspec.value_type)
-        klass = type(full_name, (base, parent), kwargs)
-
-        return self._create_type_from_pspec(pspec, klass, child=child)
-
-    def list(self, type_name, parent=None):
-        """
-        List all the properties for type_name
-        If parent is specified, child properties will also be included.
-        This method will create the prop_types and save if they're
-        not already created.
-        @param type_name: name of type
-        @type  type_name: string
-        @param parent:    parent type of object
-        @type  parent:    GType
-        """
-
-        property_types = []
-        for pspec in self._list_properties(type_name):
-            property_types.append(self._get_type_from_pspec(pspec))
-
-        if parent:
-            prop_types = []
-            for pspec in self._list_child_properties(parent):
-                prop_types.append(self._get_type_from_pspec(pspec, child=True))
-            property_types.extend(prop_types)
-
-        property_types.extend(self._list_custom(type_name))
-
-        for prop_type in property_types[:]:
-            if (not prop_type.readable or
-                not prop_type.writable or
-                prop_type.disabled):
-                property_types.remove(prop_type)
-
-        return property_types
-
-    def get_prop_type(self, type_name, prop_name, parent=None):
-        """
-        Fetch a property from type_name called prop_name, note that it
-        does not create the property type if it's not already created.
-        @param type_name: gtype name of owner
-        @type  type_name: string
-        @param prop_name: name of property
-        @type  prop_name: string
-        """
-        pspec = self._get_pspec(type_name, prop_name, parent)
-        if not pspec:
-            prop_type = self._get_custom(type_name, prop_name)
-            if not prop_type:
-                raise PropertyError("not created: %s::%s" % (type_name,
-                                                             prop_name))
-            return prop_type
-
-        child = False
-        if pspec in self._list_child_properties(parent):
-            child = True
-
-        return self._get_type_from_pspec(pspec, child=child)
-
-    def _get_type_from_pspec(self, pspec, child=False):
-        full_name = gobject.type_name(pspec.owner_type) + '::' + pspec.name
-        if full_name in self._prop_types:
-            return self._prop_types[full_name]
-
-        return self._create_type_from_pspec(pspec, child=child)
-
-    def _get_base_type(self, value_type):
-        if gobject.type_is_a(value_type, gobject.TYPE_OBJECT):
-            base_type = gobject.TYPE_OBJECT
-        elif gobject.type_is_a(value_type, gobject.TYPE_ENUM):
-            base_type = gobject.TYPE_ENUM
-        elif gobject.type_is_a(value_type, gobject.TYPE_FLAGS):
-            base_type = gobject.TYPE_FLAGS
-        elif gobject.type_is_a(value_type, gobject.TYPE_BOXED):
-            base_type = gobject.TYPE_BOXED
-        else:
-            base_type = value_type
-
-        base = self._bases.get(base_type)
-        if not base:
-            raise PropertyError("Can't find base type for %r" % base_type)
-        return base
-
-    def _create_type_from_pspec(self, pspec, base=None, child=False):
-        if not base:
-            base = self._get_base_type(pspec.value_type)
-
-        readable = pspec.flags & gobject.PARAM_READABLE != 0
-        writable = pspec.flags & gobject.PARAM_WRITABLE != 0
-        prop_type = PropMeta.new(gobject.type_name(pspec.owner_type),
-                                 pspec.name, base,
-                                 pspec.value_type,
-                                 pspec.owner_type,
-                                 readable=readable,
-                                 writable=writable,
-                                 child=child,
-                                 label=pspec.nick)
-        self._set_interval_from_pspec(prop_type, pspec)
-        self._add_type(prop_type.__name__, prop_type, pspec)
-        return prop_type
-
-    def _set_interval_from_pspec(self, prop_type, pspec):
-        if not hasattr(pspec, 'maximum'):
-            return
-
-        maximum = pspec.maximum
-        if maximum == -1 and pspec.value_type == gobject.TYPE_UINT:
-            maximum = sys.maxint
-        prop_type.maximum = maximum
-
-        prop_type.minimum = pspec.minimum
-        prop_type.defautl = pspec.default_value
-
     def _add_type(self, prop_name, prop_type, pspec=None):
         if pspec:
             if gobject.type_is_a(pspec.value_type, gobject.TYPE_ENUM):
@@ -281,18 +97,83 @@ class PropRegistry:
             if not prop_type.label:
                 prop_type.label = prop_name.capitalize()
 
-        # We can't enable this just yet. child properties
-        # are overriding existing ones, they need a another storage
-        #if prop_name in self._prop_types:
-        #    raise TypeError
+        if prop_name in self._types:
+            msg = "There is already a property called %s registered"
+            raise TypeError(msg % prop_name)
 
-        self._prop_types[prop_name] = prop_type
+        self._types[prop_name] = prop_type
 
-    def _add_custom_type(self, klass, type_name, prop_name, pspec):
-        if not type_name in self._custom:
-            custom = self._custom[type_name] = []
-        else:
-            custom = self._custom[type_name]
+    def list(self, type_name, prop_registry):
+        """
+        List all the properties for type_name
+        If parent is specified, child properties will also be included.
+        This method will create the prop_types and save if they're
+        not already created.
+
+        @param type_name: name of type
+        @type  type_name: string
+        @param prop_registry: property registry
+        @type  prop_registry: a PropRegistry
+        """
+        property_types = []
+        for pspec in self._list_properties(type_name):
+            prop = self._get_type_from_prop_name(type_name, pspec.name)
+            if not prop:
+                base = prop_registry._get_base_type(pspec.value_type)
+                prop = self.create_type_from_pspec(pspec, base, type_name)
+            property_types.append(prop)
+
+        property_types.extend(self._list_custom(type_name))
+
+        for prop_type in property_types[:]:
+            if (not prop_type.readable or
+                not prop_type.writable or
+                not prop_type.enabled):
+                property_types.remove(prop_type)
+
+        return property_types
+
+    def has_prop_type(self, type_name, prop_name):
+        """
+        True if a property from type_name called prop_name exists
+
+        @param type_name: gtype name of owner
+        @type  type_name: string
+        @param prop_name: name of property
+        @type  prop_name: string
+        """
+        prop_type = self._get_type_from_prop_name(type_name, prop_name)
+        if not prop_type:
+            prop_type = self._get_custom(type_name, prop_name)
+
+        return prop_type
+
+    def get_prop_type(self, type_name, prop_name, prop_registry):
+        """
+        Fetch a property from type_name called prop_name, it creates
+        the property type if it's not already created.
+
+        @param type_name: gtype name of owner
+        @type  type_name: string
+        @param prop_name: name of property
+        @type  prop_name: string
+        @param prop_registry: property registry
+        @type  prop_registry: a PropRegistry
+        """
+        prop_type = self._get_type_from_prop_name(type_name, prop_name)
+        if not prop_type:
+            prop_type = self._get_custom(type_name, prop_name)
+
+        if not prop_type:
+            pspec = self.get_pspec(type_name, prop_name)
+            if pspec:
+                base = prop_registry._get_base_type(pspec.value_type)
+                prop_type = self.create_type_from_pspec(pspec, base, type_name)
+
+        return prop_type
+
+    def add_custom_type(self, klass, type_name, prop_name, pspec):
+        custom = self._custom.setdefault(type_name, [])
         if prop_name in custom:
             return
         klass.custom = True
@@ -316,7 +197,7 @@ class PropRegistry:
         while True:
             for prop_name in self._custom.get(type_name, []):
                 full_name = type_name + '::' + prop_name
-                retval.append(self._prop_types[full_name])
+                retval.append(self._types[full_name])
 
             if type_name == 'GObject':
                 break
@@ -325,53 +206,271 @@ class PropRegistry:
 
         return retval
 
-    def _get_pspec(self, type_name, prop_name, parent=None):
+    def get_pspec(self, type_name, prop_name):
         for pspec in self._list_properties(type_name):
             if pspec.name == prop_name:
                 return pspec
 
-        # List child properties even if we're not specifying a
-        # parent
-        if parent:
-            type_name = parent
-
-        if gobject.type_is_a(type_name, gtk.Container):
-            for pspec in self._list_child_properties(type_name):
-                if pspec.name == prop_name:
-                    return pspec
-
-    def _has_property(self, type_name, prop_name):
+    def has_property(self, type_name, prop_name):
         "Checks if  type_name has a property with name prop_name"
-        return self._get_pspec(type_name, prop_name) != None
+        return self.get_pspec(type_name, prop_name) != None
 
     def _list_properties(self, type_name):
         "List all properties for type_name, and use a simple cache"
         pspecs = self._pspec_cache.get(type_name, [])
         if pspecs:
             return pspecs
-        else:
-            try:
+
+        try:
+            if self.child:
+                pspecs = gtk.container_class_list_child_properties(type_name)
+            else:
                 pspecs = gobject.list_properties(type_name)
-            except TypeError:
-                raise TypeError("Unknown GType name: %s" % type_name)
-            pspecs = list(pspecs)
+        except TypeError:
+            raise PropertyError("Unknown GType name: %s" % type_name)
+        pspecs = list(pspecs)
 
         self._pspec_cache[type_name] = pspecs
 
-        #try:
-        #    parent_type = gobject.type_parent(type_name)
-        #    parent_pspecs = gobject.list_properties(parent_type)
-        #except RuntimeError:
-        #    return pspecs
-
         return pspecs
 
-    def _list_child_properties(self, parent):
-        "List all the child properties for parent"
-        if not parent:
-            return []
+    def _get_type_from_prop_name(self, type_name, prop_name):
+        """Get the type for type_name::prop_name. If we don't have a PropType
+        for this type_name, check in its parent types"""
+        assert type(type_name) == str
+
+        while True:
+            full_name = type_name + '::' + prop_name
+            if full_name in self._types.keys():
+                return self._types[full_name]
+
+            if type_name == 'GObject':
+                break
+
+            type_name = gobject.type_name(gobject.type_parent(type_name))
+
+    def create_type_from_pspec(self, pspec, base, type_name=None):
+        if not type_name:
+            type_name = gobject.type_name(pspec.owner_type)
+
+        readable = pspec.flags & gobject.PARAM_READABLE != 0
+        writable = pspec.flags & gobject.PARAM_WRITABLE != 0
+        prop_type = PropMeta.new(type_name,
+                                 pspec.name,
+                                 pspec.blurb,
+                                 base,
+                                 pspec.value_type,
+                                 pspec.owner_type,
+                                 readable=readable,
+                                 writable=writable,
+                                 child=self.child,
+                                 label=pspec.nick)
+        self._set_interval_from_pspec(prop_type, pspec)
+        self._add_type(prop_type.__name__, prop_type, pspec)
+        return prop_type
+
+    def _set_interval_from_pspec(self, prop_type, pspec):
+        if not hasattr(pspec, 'maximum'):
+            return
+
+        maximum = pspec.maximum
+        if maximum == -1 and pspec.value_type == gobject.TYPE_UINT:
+            maximum = sys.maxint
+        prop_type.maximum = maximum
+
+        prop_type.minimum = pspec.minimum
+        prop_type.defautl = pspec.default_value
+
+class ChildPropStorage(PropStorage):
+    child = True
+
+class PropRegistry:
+    def __init__(self):
+        self._bases = {}
+
+        self._normal_storage = PropStorage()
+        self._child_storage = ChildPropStorage()
+
+    def add_base_type(self, klass, gtype):
+        if not issubclass(klass, PropType):
+            raise TypeError("klass needs to be a subclass of PropType")
+
+        if gtype in self._bases:
+            raise PropertyError("base type %r is already defined" % gtype)
+        elif klass.base_type != None:
+            raise PropertyError("klass type %r is already registered to %r" %
+                                (klass, klass.base_type))
+
+        self._bases[gtype] = klass
+        klass.base_type = gtype
+
+    def override_property(self, full_name, klass):
+        """
+        Create a property for full_name, (type::prop_name)
+        If the property exists, please use override_simple
+        @param klass: property class
+        @type  klass: subclass of PropType
+        @param full_name: complete name of property
+        @type  full_name: string (typename::prop_name)
+        """
+        self._override_property_internal(self._normal_storage,
+                                         full_name, klass)
+
+    def override_child_property(self, full_name, klass):
+        """
+        Create a child property for full_name, (type::prop_name)
+        If the property exists, please use override_simple_child
+        @param klass: property class
+        @type  klass: subclass of PropType
+        @param full_name: complete name of property
+        @type  full_name: string (typename::prop_name)
+        """
+        self._override_property_internal(self._child_storage, full_name, klass)
+        klass.child = True
+
+    def _override_property_internal(self, storage, full_name, klass):
+        if not issubclass(klass, PropType):
+            raise TypeError("klass needs to be a subclass of PropType")
+
+        type_name, prop_name = full_name.split('::')
+        pspec = storage.get_pspec(type_name, prop_name)
+
+        # If we can't find the property in the GType, it's a custom type
+        if not storage.has_property(type_name, prop_name):
+            storage.add_custom_type(klass, type_name, prop_name, pspec)
+        else:
+            raise TypeError("%s exists, use override_simple" % full_name)
+
+        # Note, that at this point we need to setup the overriden class
+        # since we're not going to call the type constructor, so
+        # this needs to be in sync with PropMeta.new
+
+        klass.name = prop_name
+        klass.owner_name = type_name
+        klass.owner_type = gobject.type_from_name(type_name)
+
+        if not issubclass(klass.custom_editor, PropertyCustomEditor):
+            raise TypeError(
+                "custom_editor %r for property class %r needs to be a "
+                "subclass of PropertyCustomEditor " %
+                (klass.custom_editor, klass))
+
+    def override_simple(self, full_name, base=None, **kwargs):
+        """
+        Same as override_property, but creates and registeres the class
+        @param full_name: complete name of property
+        @param base:      base class
+        @param child:     child property
+        @param kwargs namespace to use in class
+        """
+        self._override_simple_internal(self._normal_storage,
+                                       full_name, base, **kwargs)
+
+    def override_simple_child(self, full_name, base=None, **kwargs):
+        """
+        Same as override_child_property, but creates and registeres the class
+        @param full_name: complete name of property
+        @param base:      base class
+        @param child:     child property
+        @param kwargs namespace to use in class
+        """
+        self._override_simple_internal(self._child_storage,
+                                       full_name, base, **kwargs)
+
+    def _override_simple_internal(self, storage, full_name, base=None,
+                                  **kwargs):
+        type_name, prop_name = full_name.split('::')
+
+        pspec = storage.get_pspec(type_name, prop_name)
+        if not pspec:
+            raise TypeError("You can only override existing properties, %s "
+                            "is not a property" % full_name)
+
+        # check if the type already exist
+        if storage.has_prop_type(type_name, prop_name):
+            prop_type = storage.get_prop_type(type_name, prop_name, self)
+            # If a base is provided we don't modify the original proptype
+            # but create a new one. This ensures the parent prop is not changed
+            if base and prop_type == base:
+                klass = type(full_name, (base,), kwargs)
+
+                prop_type = storage.create_type_from_pspec(pspec,
+                                                           klass,
+                                                           type_name=type_name)
+            else:
+                # simply update the prop_type
+                for key, value in kwargs.items():
+                    setattr(prop_type, key, value)
+
+        else:
+            if not base:
+                base = self._get_base_type(pspec.value_type)
+
+            klass = type(full_name, (base,), kwargs)
+
+            prop_type = storage.create_type_from_pspec(pspec,
+                                                       klass,
+                                                       type_name=type_name)
+
+        return prop_type
+
+    def list(self, type_name, parent=None):
+        """
+        List all the properties for type_name
+        If parent is specified, child properties will also be included.
+        This method will create the prop_types if they're not already created.
+
+        @param type_name: name of type
+        @type  type_name: string
+        @param parent:    parent type of object
+        @type  parent:    GType
+        """
+        property_types = self._normal_storage.list(type_name, self)
+
+        if parent:
+            parent_type = gobject.type_name(parent)
+            property_types.extend(self._child_storage.list(parent_type, self))
+
+        return property_types
+
+    def get_prop_type(self, type_name, prop_name, parent=None):
+        """
+        Fetch a property from type_name called prop_name. It creates
+        the property type if it's not already created.
+
+        @param type_name: gtype name of owner
+        @type  type_name: string
+        @param prop_name: name of property
+        @type  prop_name: string
+        """
+        if parent:
+            parent_type = gobject.type_name(parent)
+            prop_type = self._child_storage.get_prop_type(parent_type,
+                                                          prop_name, self)
+
+        else:
+            prop_type = self._normal_storage.get_prop_type(type_name,
+                                                           prop_name, self)
+
+        return prop_type
+
+    def _get_base_type(self, value_type):
+        if gobject.type_is_a(value_type, gobject.TYPE_OBJECT):
+            base_type = gobject.TYPE_OBJECT
+        elif gobject.type_is_a(value_type, gobject.TYPE_ENUM):
+            base_type = gobject.TYPE_ENUM
+        elif gobject.type_is_a(value_type, gobject.TYPE_FLAGS):
+            base_type = gobject.TYPE_FLAGS
+        elif gobject.type_is_a(value_type, gobject.TYPE_BOXED):
+            base_type = gobject.TYPE_BOXED
+        else:
+            base_type = value_type
+
+        base = self._bases.get(base_type)
+        if not base:
+            raise PropertyError("Can't find base type for %r" % base_type)
+        return base
 
-        return gtk.container_class_list_child_properties(parent)
 
     def get_writable_properties(self, widget, child):
         """
@@ -381,13 +480,17 @@ class PropRegistry:
         @param widget: a gtk.Widget
         @param child: True to list child properties, False to omit them
         """
+        if child:
+            parent_type = gobject.type_name(widget.get_parent())
+            property_types = self._child_storage.list(parent_type, self)
 
-        all_properties = prop_registry.list(gobject.type_name(widget),
-                                            widget.get_parent())
+        else:
+            type_name = gobject.type_name(widget)
+            property_types = self._normal_storage.list(type_name, self)
 
         properties = []
         # write the properties
-        for prop_type in all_properties:
+        for prop_type in property_types:
             # don't save the name property since it is saved in the id
             # attribute of the tag
             if prop_type.name == 'name':
@@ -403,10 +506,6 @@ class PropRegistry:
             if not prop_type.persistent:
                 continue
 
-            # child properties are saved later
-            if prop_type.child != child:
-                continue
-
             properties.append(prop_type)
 
         return properties
@@ -419,13 +518,15 @@ class PropMeta(type):
     a factory for types which has me as a metaclass.
     """
 
-    def new(mcs, owner_name, name, base, value_type, owner_type,
+    #@classmethod
+    def new(mcs, owner_name, name, description, base, value_type, owner_type,
             readable=True, writable=True, child=False, label=None):
 
         cls = mcs(owner_name + '::' + name, (base,), {})
 
         # This can not be overridden
         cls.name =  name
+        cls.description = description
         cls.type = value_type
         cls.owner_name = owner_name
         cls.owner_type = owner_type
@@ -443,19 +544,20 @@ class PropMeta(type):
         # to take precedence over the one in the transparent subclass
         # we're going to create, so check before assigning anything.
 
-        # This is clearly wrong, some parts depend on it
-        # See widgets/base/buttons: RadioGroupProp
-        cls.readable = readable
-        cls.writable = writable
+        if cls.readable is None:
+            cls.readable = readable
 
-        if not readable:
+        if cls.writable is None:
+            cls.writable = writable
+
+        if not cls.readable:
             getter = PropType._value_not_readable
         elif child:
             getter = base._child_get
         else:
             getter = base._get
 
-        if not writable:
+        if not cls.writable:
             setter = PropType._value_not_writable
         elif child:
             setter = base._child_set
@@ -493,6 +595,7 @@ class PropType(object):
     of me wrap a GObject property.
 
     @cvar name:         name of the property, eg expand
+    @cvar description:  description of the property (it'll show in a tooltip)
     @cvar owner_name:   owner gobject name: GtkWindow
     @cvar type:         type, GType of property
     @cvar owner_type:   type of owner
@@ -509,17 +612,22 @@ class PropType(object):
     @cvar custom:
     @cvar editable:
     @cvar base_type:
-    @cvar editor:       editor instance, subclass of XXX
-    @cvar persistent:   Should we save the property?
+    @cvar editor:        editor
+    @cvar custom_editor: specialized editor
+    @cvar persistent:    Should we save the property?
+    @cvar priority:      Higher priorities appear before lower priorities in
+                         the property editor
+    @cvar has_custom_default: the prop has been customized with a default
     """
 
     name = None
+    description = None
     owner_name = None
     type = None
     owner_type = gobject.TYPE_INVALID
     value = None
-    readable = True
-    writable = True
+    readable = None
+    writable = None
     set = None
     get = None
     child = None
@@ -530,11 +638,15 @@ class PropType(object):
     custom = False
     editable = True
     base_type = None
-    editor = PropertyCustomEditor
+    editor = None
+    custom_editor = PropertyCustomEditor
     persistent = True
-    disabled = False
+    priority = 100
+    has_custom_default = False
 
     def __init__(self, gadget):
+        self._initial = None
+
         cls = type(self)
         assert cls.name != None
         assert cls.owner_name != None
@@ -555,6 +667,7 @@ class PropType(object):
 
         self.load()
 
+    #@property
     def get_object(self):
         return self.gadget.widget
     object = property(get_object)
@@ -577,7 +690,7 @@ class PropType(object):
                     self._value = parent.child_get_property(self.object,
                                                             self.name)
                 else:
-                    # FIXME: This is broke, but needed for
+                    # FIXME: This is broken, but needed for
                     # GtkRadioButton::group,
                     # see tests.test_glade.GladeTest.test_radiobutton
                     try:
@@ -585,7 +698,7 @@ class PropType(object):
                     except TypeError:
                         self._value = None
         else:
-            if self.custom:
+            if self.custom or self.has_custom_default:
                 self._initial = self.default
                 self.set(self.default)
             else:
@@ -601,6 +714,17 @@ class PropType(object):
 
         return repr(self.object)
 
+    def add_change_notify(self, function):
+        if not callable(function):
+            raise TypeError("function must be callable")
+
+        if not custom:
+            if self.child:
+                signal_name = 'child-notify'
+            else:
+                signal_name = 'notify'
+            self.object.connect('%s::%s' % (signal_name, self.name),
+                                lambda self, obj, pspec: function(self))
     def _value_not_readable(self):
         raise PropertyError("%s not readable" % self.name)
 
@@ -644,7 +768,7 @@ class PropType(object):
                 gobject.type_name(self.owner_type), self.name, value, e))
 
     def get_default(self, gobj):
-        if self.custom:
+        if self.custom or self.has_custom_default:
             return self.default
 
         if self.child:
@@ -669,8 +793,9 @@ class PropType(object):
         Return None if it already contains the default value
         """
         value = self.get()
-        if value == None or value == self._initial:
-            return
+        if not self.has_custom_default:
+            if value == None or value == self._initial:
+                return
 
         return str(value)
 
@@ -681,12 +806,25 @@ class PropType(object):
 # XXX: This is a really bad name since it conflicts with
 #      the custom widget
 class CustomProperty(PropType):
+    readable = True
+    writable = True
+    def __init__(self, object):
+        self._functions = []
+        super(CustomProperty, self).__init__(object)
+
     def get(self):
         return self._get()
 
     def set(self, value):
         self._set(value)
 
+    def add_change_notify(self, function):
+        self._functions.append(function)
+
+    def notify(self):
+        for function in self._functions:
+            function(self)
+
     value = property(lambda self: self.get(),
                      lambda self, value: self.set(value))
 
@@ -698,6 +836,10 @@ if environ.epydoc:
             pass
         def override_property(self, *args, **kwargs):
             pass
+        def override_simple(self, *args, **kwargs):
+            pass
+        def get_prop_type(self, *args, **kwargs):
+            pass
     prop_registry = FakeRegistry()
     custom = object
 else:
@@ -780,6 +922,7 @@ class BaseNumericType(PropType):
     # number of decimals
     digits = 0
 
+    #@classmethod
     def get_adjustment(cls):
         #if not cls.minimum <= cls.default <= cls.maximum:
         #    raise TypeError(
@@ -803,6 +946,17 @@ class BaseFloatType(BaseNumericType):
     minimum = -10.0**308
     maximum = 10.0**308
     digits = 2
+    def save(self):
+        if not BaseNumericType.save(self):
+            return
+        # Fix for #326723
+        # kiwi/datatypes.py uses %.12g here, but it seems to
+        # be too precise, 0.1 -> 0.10000000149
+        value = self.get()
+        strvalue = str(value)
+        if strvalue.endswith('.0'):
+            return strvalue
+        return '%.8g' % value
 
 class IntType(BaseNumericType):
     minimum = -int(2**31-1)
@@ -828,6 +982,8 @@ prop_registry.add_base_type(DoubleType, gobject.TYPE_DOUBLE)
 
 class EnumType(PropType):
     enum_class = None
+
+    #@classmethod
     def get_choices(cls):
         choices = []
         if not hasattr(cls.enum_class, '__enum_values__'):
@@ -850,6 +1006,8 @@ prop_registry.add_base_type(EnumType, gobject.TYPE_ENUM)
 
 class FlagsType(PropType):
     flags_class = None
+
+    #@classmethod
     def get_choices(cls):
         if not hasattr(cls.flags_class, '__flags_values__'):
             raise UnsupportedProperty
@@ -864,7 +1022,7 @@ class FlagsType(PropType):
 
     def save(self):
         value = self.get()
-        if value == self._initial or value == 0:
+        if value == self._initial:
             return
         return flags_to_string(value, flags=self.flags_class)
 
@@ -915,6 +1073,59 @@ class ProxiedProperty(TransparentProperty):
 #   TYPE_ULONG
 #   TYPE_UNICHAR
 
+# Property commands
+class CommandSetProperty(Command):
+    def __init__(self,  property, value):
+        description = _('Setting %s of %s') % (
+            property.name, property.get_object_name())
+        Command.__init__(self, description)
+        self._property = property
+        self._value = value
+
+    def execute(self):
+        new_value = self._value
+        # store the current value for undo
+        self._value = self._property.value
+        self._property.set(new_value)
+
+        # TODO: this should not be needed
+        # if the property is the name, we explicitily set the name of the
+        # gadget to trigger the notify signal so several parts of the
+        # interface get updated
+        if self._property.name == 'name':
+            self._property.object.notify('name')
+
+    def unifies(self, other):
+        if isinstance(other, CommandSetProperty):
+            return self._property is other._property
+        return False
+
+    def collapse(self, other):
+        self.description = other.description
+        other.description = None
+
+class CommandSetTranslatableProperty(CommandSetProperty):
+    def __init__(self,  property, value, comment, translatable, has_context):
+        CommandSetProperty.__init__(self, property, value)
+        self._comment = comment
+        self._translatable = translatable
+        self._has_context = has_context
+
+    def execute(self):
+        new_comment = self._comment
+        new_translatable = self._translatable
+        new_has_context = self._has_context
+
+        # store the current value for undo
+        self._comment = self._property.i18n_comment
+        self._translatable = self._property.translatable
+        self._has_context = self._property.has_i18n_context
+
+        CommandSetProperty.execute(self)
+
+        self._property.translatable = new_translatable
+        self._property.i18n_comment = new_comment
+        self._property.has_i18n_context = new_has_context
 
 if __name__ == '__main__':
     win = gtk.Window()
